@@ -55,9 +55,9 @@ default_settings = {
     'sectsubtitle_xform': False,
 }
 
-# This is increased every time a new environment attribute is added
-# to properly invalidate pickle files.
-ENV_VERSION = 18
+# This is increased every time an environment attribute is added
+# or changed to properly invalidate pickle files.
+ENV_VERSION = 21
 
 
 def walk_depth(node, depth, maxdepth):
@@ -218,8 +218,10 @@ class BuildEnvironment:
         # All "docnames" here are /-separated and relative and exclude the source suffix.
 
         self.found_docs = set()     # contains all existing docnames
-        self.all_docs = {}          # docname -> (mtime, md5sum) at the time of build
+        self.all_docs = {}          # docname -> mtime at the time of build
                                     # contains all built docnames
+        self.dependencies = {}      # docname -> set of dependent file names, relative to
+                                    # documentation root
 
         # File metadata
         self.metadata = {}          # docname -> dict of metadata items
@@ -249,6 +251,7 @@ class BuildEnvironment:
                                     # (type, string, target, aliasname)
         self.versionchanges = {}    # version -> list of
                                     # (type, docname, lineno, module, descname, content)
+        self.images = {}            # absolute path -> (docnames, unique filename)
 
         # These are set while parsing a file
         self.docname = None         # current document name
@@ -267,9 +270,11 @@ class BuildEnvironment:
         self._warnfunc = func
         self.settings['warning_stream'] = RedirStream(func)
 
-    def warn(self, docname, msg):
+    def warn(self, docname, msg, lineno=None):
         if docname:
-            self._warnfunc(self.doc2path(docname) + ':: ' + msg)
+            if lineno is None:
+                lineno = ''
+            self._warnfunc('%s:%s: %s' % (self.doc2path(docname), lineno, msg))
         else:
             self._warnfunc('GLOBAL:: ' + msg)
 
@@ -278,16 +283,18 @@ class BuildEnvironment:
         if docname in self.all_docs:
             self.all_docs.pop(docname, None)
             self.metadata.pop(docname, None)
+            self.dependencies.pop(docname, None)
             self.titles.pop(docname, None)
             self.tocs.pop(docname, None)
             self.toc_num_entries.pop(docname, None)
+            self.filemodules.pop(docname, None)
+            self.indexentries.pop(docname, None)
 
             for subfn, fnset in self.files_to_rebuild.iteritems():
                 fnset.discard(docname)
             for fullname, (fn, _) in self.descrefs.items():
                 if fn == docname:
                     del self.descrefs[fullname]
-            self.filemodules.pop(docname, None)
             for modname, (fn, _, _, _) in self.modules.items():
                 if fn == docname:
                     del self.modules[modname]
@@ -297,10 +304,13 @@ class BuildEnvironment:
             for key, (fn, _) in self.reftargets.items():
                 if fn == docname:
                     del self.reftargets[key]
-            self.indexentries.pop(docname, None)
             for version, changes in self.versionchanges.items():
                 new = [change for change in changes if change[1] != docname]
                 changes[:] = new
+            for fullpath, (docs, _) in self.images.items():
+                docs.discard(docname)
+                if not docs:
+                    del self.images[fullpath]
 
     def doc2path(self, docname, base=True, suffix=None):
         """
@@ -318,14 +328,18 @@ class BuildEnvironment:
         else:
             return path.join(base, docname.replace(SEP, path.sep)) + suffix
 
-    def get_outdated_files(self, config, config_changed):
+    def find_files(self, config):
         """
-        Return (added, changed, removed) sets.
+        Find all source files in the source dir and put them in self.found_docs.
         """
         self.found_docs = set(get_matching_docs(self.srcdir, config.source_suffix,
                                                 exclude=set(config.unused_docs),
                                                 prune=['_sources']))
 
+    def get_outdated_files(self, config_changed):
+        """
+        Return (added, changed, removed) sets.
+        """
         # clear all files no longer present
         removed = set(self.all_docs) - self.found_docs
 
@@ -339,17 +353,34 @@ class BuildEnvironment:
             for docname in self.found_docs:
                 if docname not in self.all_docs:
                     added.add(docname)
-                else:
-                    # if the doctree file is not there, rebuild
-                    if not path.isfile(self.doc2path(docname, self.doctreedir,
-                                                     '.doctree')):
-                        changed.add(docname)
-                        continue
-                    mtime, md5sum = self.all_docs[docname]
-                    newmtime = path.getmtime(self.doc2path(docname))
-                    if newmtime == mtime:
-                        continue
+                    continue
+                # if the doctree file is not there, rebuild
+                if not path.isfile(self.doc2path(docname, self.doctreedir,
+                                                 '.doctree')):
                     changed.add(docname)
+                    continue
+                # check the mtime of the document
+                mtime = self.all_docs[docname]
+                newmtime = path.getmtime(self.doc2path(docname))
+                if newmtime > mtime:
+                    changed.add(docname)
+                    continue
+                # finally, check the mtime of dependencies
+                for dep in self.dependencies.get(docname, ()):
+                    try:
+                        # this will do the right thing when dep is absolute too
+                        deppath = path.join(self.srcdir, dep)
+                        if not path.isfile(deppath):
+                            changed.add(docname)
+                            break
+                        depmtime = path.getmtime(deppath)
+                        if depmtime > mtime:
+                            changed.add(docname)
+                            break
+                    except EnvironmentError:
+                        # give it another chance
+                        changed.add(docname)
+                        break
 
         return added, changed, removed
 
@@ -369,12 +400,14 @@ class BuildEnvironment:
                     continue
                 if not hasattr(self.config, key) or \
                    self.config[key] != config[key]:
+
                     msg = '[config changed] '
                     config_changed = True
                     break
             else:
                 msg = ''
-        added, changed, removed = self.get_outdated_files(config, config_changed)
+        self.find_files(config)
+        added, changed, removed = self.get_outdated_files(config_changed)
         msg += '%s added, %s changed, %s removed' % (len(added), len(changed),
                                                      len(removed))
         yield msg
@@ -394,6 +427,12 @@ class BuildEnvironment:
             self.warn(None, 'master file %s not found' %
                       self.doc2path(config.master_doc))
 
+        # remove all non-existing images from inventory
+        for imgsrc in self.images.keys():
+            if not os.access(path.join(self.srcdir, imgsrc), os.R_OK):
+                del self.images[imgsrc]
+
+
     # --------- SINGLE FILE BUILDING -------------------------------------------
 
     def read_doc(self, docname, src_path=None, save_parsed=True, app=None):
@@ -409,18 +448,15 @@ class BuildEnvironment:
         doctree = publish_doctree(None, src_path, FileInput,
                                   settings_overrides=self.settings,
                                   reader=MyStandaloneReader())
+        self.process_dependencies(docname, doctree)
+        self.process_images(docname, doctree)
         self.process_metadata(docname, doctree)
         self.create_title_from(docname, doctree)
         self.note_labels_from(docname, doctree)
         self.build_toc_from(docname, doctree)
 
-        # calculate the MD5 of the file at time of build
-        f = open(src_path, 'rb')
-        try:
-            md5sum = md5(f.read()).digest()
-        finally:
-            f.close()
-        self.all_docs[docname] = (path.getmtime(src_path), md5sum)
+        # store time of reading, used to find outdated files
+        self.all_docs[docname] = time.time()
 
         if app:
             app.emit('doctree-read', doctree)
@@ -430,6 +466,7 @@ class BuildEnvironment:
         doctree.transformer = None
         doctree.settings.warning_stream = None
         doctree.settings.env = None
+        doctree.settings.record_dependencies = None
 
         # cleanup
         self.docname = None
@@ -452,6 +489,46 @@ class BuildEnvironment:
         else:
             return doctree
 
+    def process_dependencies(self, docname, doctree):
+        """
+        Process docutils-generated dependency info.
+        """
+        deps = doctree.settings.record_dependencies
+        if not deps:
+            return
+        docdir = path.dirname(self.doc2path(docname, base=None))
+        for dep in deps.list:
+            dep = path.join(docdir, dep)
+            self.dependencies.setdefault(docname, set()).add(dep)
+
+    def process_images(self, docname, doctree):
+        """
+        Process and rewrite image URIs.
+        """
+        existing_names = set(v[1] for v in self.images.itervalues())
+        docdir = path.dirname(self.doc2path(docname, base=None))
+        for node in doctree.traverse(nodes.image):
+            imguri = node['uri']
+            if imguri.find('://') != -1:
+                self.warn(docname, 'Nonlocal image URI found: %s' % imguri, node.line)
+            else:
+                imgpath = path.normpath(path.join(docdir, imguri))
+                node['uri'] = imgpath
+                self.dependencies.setdefault(docname, set()).add(imgpath)
+                if not os.access(path.join(self.srcdir, imgpath), os.R_OK):
+                    self.warn(docname, 'Image file not readable: %s' % imguri, node.line)
+                if imgpath in self.images:
+                    self.images[imgpath][0].add(docname)
+                    continue
+                uniquename = path.basename(imgpath)
+                base, ext = path.splitext(uniquename)
+                i = 0
+                while uniquename in existing_names:
+                    i += 1
+                    uniquename = '%s%s%s' % (base, i, ext)
+                self.images[imgpath] = (set([docname]), uniquename)
+                existing_names.add(uniquename)
+
     def process_metadata(self, docname, doctree):
         """
         Process the docinfo part of the doctree as metadata.
@@ -492,6 +569,8 @@ class BuildEnvironment:
             if not explicit:
                 continue
             labelid = document.nameids[name]
+            if labelid is None:
+                continue
             node = document.ids[labelid]
             if name.isdigit() or node.has_key('refuri') or \
                    node.tagname.startswith('desc_'):
@@ -500,7 +579,8 @@ class BuildEnvironment:
                 continue
             if name in self.labels:
                 self.warn(docname, 'duplicate label %s, ' % name +
-                          'other instance in %s' % self.doc2path(self.labels[name][0]))
+                          'other instance in %s' % self.doc2path(self.labels[name][0]),
+                          node.line)
             self.anonlabels[name] = docname, labelid
             if not isinstance(node, nodes.section):
                 # anonymous-only labels
@@ -581,11 +661,12 @@ class BuildEnvironment:
     # -------
     # these are called from docutils directives and therefore use self.docname
     #
-    def note_descref(self, fullname, desctype):
+    def note_descref(self, fullname, desctype, line):
         if fullname in self.descrefs:
             self.warn(self.docname,
                       'duplicate canonical description name %s, ' % fullname +
-                      'other instance in %s' % self.doc2path(self.descrefs[fullname][0]))
+                      'other instance in %s' % self.doc2path(self.descrefs[fullname][0]),
+                      line)
         self.descrefs[fullname] = (self.docname, desctype)
 
     def note_module(self, modname, synopsis, platform, deprecated):
@@ -602,6 +683,12 @@ class BuildEnvironment:
     def note_versionchange(self, type, version, node, lineno):
         self.versionchanges.setdefault(version, []).append(
             (type, self.docname, lineno, self.currmodule, self.currdesc, node.astext()))
+
+    def note_dependency(self, filename):
+        basename = path.dirname(self.doc2path(self.docname, base=None))
+        # this will do the right thing when filename is absolute too
+        filename = path.join(basename, filename)
+        self.dependencies.setdefault(self.docname, set()).add(filename)
     # -------
 
     # --------- RESOLVING REFERENCES AND TOCTREES ------------------------------
@@ -739,7 +826,8 @@ class BuildEnvironment:
                     docname, labelid = self.reftargets.get((typ, target), ('', ''))
                     if not docname:
                         if typ == 'term':
-                            self.warn(fromdocname, 'term not in glossary: %s' % target)
+                            self.warn(fromdocname, 'term not in glossary: %s' % target,
+                                      node.line)
                         newnode = contnode
                     else:
                         newnode = nodes.reference('', '')
